#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/stat.h>

#include <assert.h>
#include <fcntl.h>
#include "socket_wrapper.h"

#ifdef USE_SENDFILE
#include <sys/sendfile.h>
#ifndef DEFAULT_SEND_FILE_CHUNK_SIZE
const size_t SEND_FILE_CHUNK_SIZE = 0x100000; /*1MB*/
#else
const size_t SEND_FILE_CHUNK_SIZE = DEFAULT_SEND_FILE_CHUNK_SIZE; /*1MB*/
#endif
#endif
#ifndef DEFAULT_TIMEOUT
static const int TIMEOUT = 5;
#else
static const int TIMEOUT = DEFAULT_TIMEOUT;
#endif
/*========
 *Operation
 *========*/

int sendReadOp(int sock,const char * filename){
    struct ReadOp op;
    op.file_url_size = strlen(filename);
    op.padding0 = 0;
    if(send(sock,&op,sizeof(op),0)<0){
        perror("readop send fail");
        return -1;
    }
    if(send(sock,filename,op.file_url_size,0)<0){
        perror("readop filename send fail");
        return -1;
    }
    return 0;
}

int recvFile(int sock, const char * filename,size_t file_size){
    int fd;
    size_t count = 0;
    int i;
    int cur_progress = 1;
    int return_value = 0;
    int buf_sz = getBufferSizeFrom(sock);
    uint8_t * buf = malloc(buf_sz*sizeof(*buf));
    if (buf == NULL){
        return -1;
    }
    fd = open(filename,O_WRONLY|O_CREAT|O_TRUNC,S_IRWXU|S_IRWXG|S_IROTH);
    if (fd < 0)
    {
        perror("file open fail");
        return -1;
    }
    
    while (file_size - count > 0)
    {
        int readed = buf_sz < file_size - count ? buf_sz : file_size - count;
        if((i = recv_until_byte(sock,buf,readed,TIMEOUT)) < 0){
            if(i == -2) 
                fprintf(stderr,"recv file failed : timeout connetion lost\n");
            else perror("recv file failed");
            return_value = -1;
            goto END;
        }
        if(write(fd,buf,readed)<0){
            perror("file write failed");
            return_value = -1;
            goto END;
        }
        if( ((double)count / (double)file_size) * 100.0 > ((double)cur_progress) ){
            printf("\rprogress : %d%% current bytes: %ld bytes",cur_progress,count);
            cur_progress = (int)((((double)count / (double)file_size)) * 100.0 + 1.0);
            fflush(stdout);
        }
        count += readed;
    }
    printf("\rprogress : 100.00%% current bytes: %ld bytes\n",count);
END:
    free(buf);
    return return_value;
}

int recvData(int sock,const char * filename){
    struct TransferResult res;
    int i=0;
    if((i=recv_until_byte(sock,&res,sizeof(res),TIMEOUT)) < 0){
        if (i == -2) fprintf(stderr,"timeout");
        else perror("recv fail");
        return -1;
    }
    static char error_meesage_buf[80] = "";
    switch(res.res){
        case RES_ERR:
            fprintf(stderr,"Server Fail: %s", strerror(res.err_number));
            return -1;
        case RES_USR_ERR:
            assert(res.error_msg_size < 80);/*todo : fix*/
            if((i=recv_until_byte(sock,error_meesage_buf,res.error_msg_size,TIMEOUT)) < 0){
                if (i == -2) fprintf(stderr,"timeout");
                else perror("recv fail");
                return -1;
            }
            fprintf(stderr,"Error Message From Server: %s",error_meesage_buf);
            return -1;
        case RES_OK:
            return recvFile(sock,filename,res.file_size);
            break;
        default:
            fprintf(stderr,"unknown value!");
            return -1;
    }

    return 0;
}

int main(int argc, const char *argv[]){
    struct sockaddr_in addr;
    const char * filename;
    const char * server_name;
    in_port_t server_port = 0;
    int sock, err;
    if (argc != 4){
        fprintf(stderr,"invaild arguments number.");
        return 1;
    }
    server_name = argv[1];
    server_port = atoi(argv[2]);
    filename = argv[3];
    if (server_port == 0)
    {
        fprintf(stderr,"port invalid");
        return 1;
    }
    sock = socket(AF_INET,SOCK_STREAM,0);
    if(sock < 0){
        perror("sock create fail");
        return 1;
    }
    err = getsockaddrbyname(AF_INET,SOCK_STREAM,0,server_name,(struct sockaddr *)&addr);
    if (err != 0){
        int check;
        fprintf(stderr,"netdb fail: %s\n",gai_strerror(err));
        //assume that sernmae is *.*.*.* and try to parse addr
        check = inet_pton(AF_INET,server_name,&addr.sin_addr);
        assert(check != -1);
        if (check == 0){
            fprintf(stderr,"parsing fail : invaild format\n");
            close(sock);
            return 1;
        }
    }
    addr.sin_family = AF_INET;
    addr.sin_port = htons(server_port);
    if(connect(sock,(struct sockaddr *)&addr,sizeof(addr)) < 0){
        perror("connect failed");
        return 1;
    }
    if(sendReadOp(sock,filename) == 0){
        int ret = recvData(sock,filename);
        close(sock);
        return ret;
    }
    close(sock);
    return 0;
}